/* 
 * THC-Shagg - A modular application to bruteforce check digit algorithms
 * (c) 2001 Plasmoid <plasmoid@thehackerschoice.com>                      
 * The Hacker's Choice, http://www.thehackerschoice.com
 * $Id: Laboratory.java,v 1.2 2001/12/14 14:13:59 plasmoid Exp $ 
 */
package com.thehackerschoice.shagg;
import com.thehackerschoice.shagg.algorithms.*;
import com.thehackerschoice.shagg.util.*;

import java.util.Vector;
import java.util.TreeMap;
import java.io.*;

/**
 * This class is the laboratoy of THC-Shagg. The provided filename is
 * loaded and parsed. The set of of retrieved serial numbers gets brute-
 * forced with all implemented check digit algorithms, weights and 
 * partitions. 
 *
 * @see com.thehackerschoice.shagg.algorithms.CheckDigitAlgorithm
 *
 * @version $Id: Laboratory.java,v 1.2 2001/12/14 14:13:59 plasmoid Exp $
 * @author Plasmoid (plasmoid@thehackerschoice.com), 
 *                  <a href="http://www.thehackerschoice.com">THC</a>
 */
public class Laboratory
{
   /* 
    * Object variables.
    */
   private CheckDigitAlgorithm[] algorithms;
   private WeightSet weightSet;
   private PartitionSet partitionSet;
   private int[][] serialNumbers;

   /** 
    * Construct a shaggadalic object using the serial numbers from 
    * the provided file. If the file doesn't contain serial numbers
    * or numbers of mixed type, Shagg will abort.
    * @param filename File containing serial numbers of the same type. 
    */
   public Laboratory(String filename) 
   {
      serialNumbers=readSerialNumberFile(filename);
      Shagg.serialNumberLen=serialNumbers[0].length;
      
         /*
          * If the serial number is shorter than the minimum partition
          * set the minimum partition length to the serial number length.
          */
      if(Shagg.serialNumberLen<Shagg.minPartitionLen)
         Shagg.minPartitionLen=Shagg.serialNumberLen;
      
      Shagg.print("Initializing algorithms:  ");
      algorithms=Shagg.initAlgorithms();
      Shagg.println("Done.");
      
      Shagg.print("Calculating weights:      ");
      weightSet=new WeightSet(Shagg.digitRadix,Shagg.maxWeightsLen);
      Shagg.println("Done.");
      
      Shagg.print("Calculating partitions:   ");
      partitionSet=new PartitionSet(Shagg.serialNumberLen, 
                                    Shagg.minPartitionLen);
      Shagg.println("Done.");
      
      /*
       * Calculate how many algorithm uses weights.
       */
      int weighted=0;
      for(int i=0;i<algorithms.length;i++)
         if(algorithms[i].usesWeights())
            weighted++;
            
      /*
       * Total states, distinguishing between weighted and non-weighted
       * algorithms.
       */
      long tmp=serialNumbers.length * partitionSet.size() *
               weighted * weightSet.size() + 
               serialNumbers.length * partitionSet.size() * 
               (algorithms.length-weighted) * 1;
      
      Shagg.println("Total states to analyse:  "+tmp);
   }

   /**
    * Read all numbers from the given file.
    * This method looks for serial numbers in each line of the file. Non-
    * digits are ignored and comments can be added to the file using the
    * common tags "#" and ";". 
    * The length of the serial numbers is determined automatically. In 
    * case the user has already specified a length, it is check against
    * the automatic one.
    * @param filename Filename containing serial numbers. 
    * @return Array of serial numbers.
    */
   private int[][] readSerialNumberFile(String filename) {
      Vector v;
      StringBuffer num; 
      BufferedReader br; 
      int[][] serialNumbers;
      int length;

      v=new Vector();   

      try {   
         br = new BufferedReader(new FileReader(filename));
      
         while(br.ready()) {
            num=new StringBuffer(br.readLine());

            for(int i=0;i<num.length();i++) { 
               
               /*
                * These are comments, everything behind those is ignored
                */
               if(num.charAt(i)==';' || num.charAt(i)=='#') {
                  num.delete(i,num.length());
                  continue;
               }
               
               /*
                * Delete all non-digits.
                */ 
               if(num.charAt(i)<'0' || num.charAt(i)>'9')
                  num.deleteCharAt(i--);
            }      

            if(num.length()>0)
               v.add(num);     
         }
      } catch(IOException e) {
         Shagg.abort(e.getMessage());
      }   
      
      if(v.size()==0)
         Shagg.abort(filename+" (No valid serial numbers found)");
     
      length=((StringBuffer)v.get(0)).length(); 
      serialNumbers=new int[v.size()][length];
      
      /*
       * We don't operate on a small set of serial numbers. 
       */
      if(v.size()<Shagg.minSerialNumbers)
         Shagg.abort(filename+" (At least "+
                     Shagg.minSerialNumbers+" serial numbers needed)");
      
      for(int i=0;i<v.size();i++) {
         num=(StringBuffer)v.get(i);
         if(num.length()!=length)
            Shagg.abort(filename+" (Numbers don't have a unique length)");

         if(Shagg.serialNumberLen!=0 && num.length()!=Shagg.serialNumberLen)
            Shagg.abort(filename+" (Numbers don't have specified length)");

         for(int j=0;j<length;j++)
            serialNumbers[i][j]=Character.getNumericValue(num.charAt(j));
      }
      
      return serialNumbers;
   }
   
   /**
    * Bruteforce through the serial numbers. 
    * This is the laboratories work horse. As described above every
    * algorithm, weight, partition is checked against every number.  Every
    * combination that matches is called a "state" and stored in a
    * red-black-tree (can be any databstructure with efficient insertion and
    * searching, e.g. O(lg n).
    * <p>
    * After collecting all states, states that donnot match all serial number
    * (failure tolerance may change this) are removed. In case the collection
    * gets to be before the end, an emergency clear is issued.
    */
   public void work() {
      int[] serialNumber, weights, partition;
      double num, currNum, currInd, percent;
      Vector results;
      TreeMap found;
      State state;
      int n,a,w,ws,p;
      boolean cleaned=false;
   
      found=new TreeMap();
      num=serialNumbers.length*algorithms.length*partitionSet.size();
      currNum=0;
      
      /*
       * All serial numbers ...
       */
      for(n=0;n<serialNumbers.length;n++) {
      
         serialNumber=serialNumbers[n];
      
         Shagg.print("\n[*] Analyzing serial number ");
         for(int i=0;i<Shagg.serialNumberLen;i++)
            Shagg.print(serialNumber[i]+"");
         Shagg.print("\n\t");
         currInd=0;
        
         /*
          * All partitions ...
          */ 
         for(p=0;p<partitionSet.size();p++) {
            partition=partitionSet.get(p);
            
            /*
             * All check digit algorithms ...
             */
            for(a=0;a<algorithms.length;a++) {
               
               /*
                * Uses weights only if necessary, some algorithms don't use
                * weights, in this case we can save some time.
                */ 
               if(algorithms[a].usesWeights()) {
                  
                  /*
                   * All weights ...
                   */
                  for(w=0;w<weightSet.size();w++) {
                     weights=weightSet.get(w);
                     state=new State(weights,a,partition,
                                     serialNumbers.length);

                     if(state.check(serialNumber,algorithms)) 
                        state.hit(found);
                  }    
               } else { 
                  state=new State(null,a,partition,serialNumbers.length);
                  if(state.check(serialNumber,algorithms)) 
                     state.hit(found);
               }

               if(currInd%50==49 || currNum==num) {
                  percent=Math.round(currNum/num*100);
                  Shagg.print(" [");
                  if(percent<10) 
                     Shagg.print("  ");
                  else if(percent<100)
                     Shagg.print(" ");
                  Shagg.print((int)percent+"%]\n\t");
               } else {
                 Shagg.print(".");
               }
               currInd++;
               currNum++;               
               
               /*
                * In case found.size() is bigger than maxStatesUncleared we
                * need to clear the collection in order to reduce memory
                * usage, that can be really extreme at this point of
                * analysis.
                */
               if(found.size()>Shagg.maxStatesUncleared) {
               
                  if(n-Shagg.failureTolerance<=0 || cleaned) 
                     Shagg.abort("Too many states, try reducing partition "+ 
                                 "minimum");
               
                  Shagg.print("\n[-] Pre-clearing states: ");
                  results=State.clear(found,n-Shagg.failureTolerance);
                  Shagg.print("Done.\n");
                  Shagg.print("[+] Re-adding states: ");
                  found=new TreeMap();
                  for(int i=0;i<results.size();i++) {
                     state=(State)results.get(i);
                     found.put(state.getId(),state);
                  }   
                  
                  Shagg.print("Done.\n\t");
                  cleaned=true;
                  currInd=0;
               } else {
                  cleaned=false;
               }
            }   
         }
      }
      
      Shagg.println("\n");
      Shagg.print("Clearing states:   ");
      results=(Vector)State.clear(found,n-Shagg.failureTolerance);      
      Shagg.println("Done.");
      Shagg.println("States matching:   "+results.size()+"\n");


      /*
       * If requested print found states to STDOUT.
       */  
      if(Shagg.verbose) {
         if(results.size()<Shagg.maxPrintedStates) {
            for(int i=0;i<results.size();i++) 
               ((State)results.get(i)).print(algorithms);
         } else {
            System.out.println("Too many states match, use more serial numbers");   
         }   
      }  
      
      /*
       * If requested save states to a file
       */
      if(Shagg.saveFilename!=null && 
         results.size() < Shagg.maxPrintedStates) {
         Shagg.print("Saving states to "+Shagg.saveFilename+": ");
         
         try {
            State.saveVector(Shagg.saveFilename,results);
         } catch(IOException e) {
            Shagg.abort(e.getMessage());
         }
         
         Shagg.println("Done");
      } 
   }
}


