/* 
 * THC-Shagg - A modular application to bruteforce check digit algorithms
 * (c) 2001 Plasmoid <plasmoid@thehackerschoice.com>                      
 * The Hacker's Choice, http://www.thehackerschoice.com
 * $Id: Shagg.java,v 1.11 2001/12/14 14:33:07 plasmoid Exp $ 
 */
package com.thehackerschoice.shagg;
import com.thehackerschoice.shagg.algorithms.*;
import com.thehackerschoice.shagg.util.*;

import java.util.Vector;
import java.io.*;

/**
 * THC-Shagg is a modular application to break serial numbers that use known
 * check digit algorithms. A supplied set of serial numbers is checked
 * against all algorithms in shagg.algorithms using all possible partitions
 * and weights (if the algorithm uses weights).
 * <p>
 * Each combination is stored as a "state" and maintained in a red-black
 * tree (using the Java implementation TreeMap, this may change in future
 * versions).  If all numbers have been analysised all states are removed
 * that donnot match a specified number of serial numbers, by default they
 * need to match all numbers. In case of a large set of serial numbers, an
 * automatic clear is forced during the progress to reduce memory usage and
 * insertion time.
 * <p>
 * If a specific number of states has have been found these states can be
 * saved to a file. Using this file THC-Shagg is able to generate more 
 * serial numbers that match the states and have the same properties as the 
 * initial set of serial numbers.
 * <p>
 * THC-Shagg is the future of THC-Credit and THC-PassId. 
 * <p>
 * The analyse is done inside the class Laboratory and the generation of 
 * new serial numbers is done in the class Factory. Simple, he?
 *
 * @see com.thehackerschoice.shagg.Laboratory
 * @see com.thehackerschoice.shagg.Factory 
 * @see com.thehackerschoice.shagg.algorithms.CheckDigitAlgorithm
 *
 * @version $Id: Shagg.java,v 1.11 2001/12/14 14:33:07 plasmoid Exp $
 * @author Plasmoid (plasmoid@thehackerschoice.com), 
 *                  <a href="http://www.thehackerschoice.com">THC</a>
 */
public class Shagg 
{
   /*
    * Some constants.
    */
   private static final String README_FILE="README";
   private static final String USAGE_FILE="USAGE";
   private static final String VERSION_FILE="VERSION";

   protected static final int minSerialNumbers=3;
   protected static final int digitRadix=10; 

   /*
    * Variables defined by commandline options or from the environment
    */
   protected static int serialNumberLen=0; 
   protected static int maxWeightsLen=3;
   protected static int minPartitionLen=7;
   protected static int maxPrintedStates=20;
   protected static int failureTolerance=0;
   protected static int numbersToGenerate=10;
   protected static long maxStatesUncleared=100000;
   protected static boolean verbose=true; 
   protected static String saveFilename=null; 
   protected static int[] numberPrefix = new int[0];  

   private static boolean generationMode=false; 
   private static boolean listStates=false;
   private static Class shagg;

   public static CheckDigitAlgorithm[] initAlgorithms() {
      CheckDigitAlgorithm[] algorithms = new CheckDigitAlgorithm[6]; 

      algorithms[0]=new Modulus10AlgorithmBasic();
      algorithms[1]=new Modulus10AlgorithmLuhn();      
      algorithms[2]=new Modulus10AlgorithmCC();
      algorithms[3]=new Modulus11Algorithm();
      algorithms[4]=new Modulus9Algorithm();
      algorithms[5]=new Modulus7Algorithm();
  
      return algorithms;
   }

   public static void abort(String message) {
      System.err.println("Error: "+message);
      System.exit(0);
   }

   public static void print(String message) {
      if(verbose)
        System.out.print(message);
   }
   
   public static void println(String message) {
      if(verbose)
         System.out.println(message);
   }
 
   public static void printFile(String filename) {
      try {
         BufferedReader br = new BufferedReader(new InputStreamReader(
                                 shagg.getResource(filename).openStream()));
         while(br.ready()) 
            System.out.println(br.readLine());
      } catch (Exception e) {
        abort("File broken. Could not read "+filename+" file.");
      }
   }

   public static void printAlgorithms() {
      CheckDigitAlgorithm[] a = initAlgorithms();
      System.out.println("\n Implemented check digit algorithms: ");
      for(int i=0;i<a.length;i++) {
         System.out.println(" [*] "+a[i].getName());
         System.out.print("\tSymbol:  "+a[i].getSymbol());
         System.out.print("\tWeights: "+a[i].usesWeights());
         System.out.println("\tDigit in number: "
            +(a[i].getCheckDigitPosition(1)>=0));
      }   
      System.out.println();
   }

   public static void printStates(String filename) {
      Vector results=null; 
      
      try {
         results=State.loadVector(filename);
      } catch (Exception e) {
         abort("Could not list states, "+filename+" is invalid.");
      }
      
      System.out.println("\nStates found in "+filename+": \n");
      for(int i=0;i<results.size();i++) 
         ((State)results.get(i)).print(initAlgorithms());
   }

   public static void main(String[] args) throws IOException {

      try {
         shagg=Class.forName("com.thehackerschoice.shagg.Shagg");
      } catch (Exception e) {
         abort("File broken. Need to locate my own class.");
      }
      
      int c;
      Getopt g = new Getopt("Error", args, "c:d:ghi:l:n:p:qr:s:w:ALP:RV");
      
      while((c=g.getopt())!=-1) {
         switch(c) {
            case 'g':
               generationMode=true;
               break;
            case 'q':
               verbose=false;
               break;
            case 'h':
               printFile(USAGE_FILE);
               System.exit(0);
               break;
            case 'p':
               minPartitionLen=Integer.parseInt(g.getOptarg());
               break;
            case 'w':
               maxWeightsLen=Integer.parseInt(g.getOptarg());
               break;
            case 'l':
               serialNumberLen=Integer.parseInt(g.getOptarg());
               break;      
            case 'i':
               failureTolerance=Integer.parseInt(g.getOptarg());
               break;
            case 'd':
               maxPrintedStates=Integer.parseInt(g.getOptarg());
               break;       
            case 'c':
               maxStatesUncleared=Long.parseLong(g.getOptarg());                           
               break;
            case 's':
               saveFilename=g.getOptarg();
               break;
            case 'P':
               String tmp=g.getOptarg();
               numberPrefix=new int[tmp.length()];
               for(int j=0;j<tmp.length();j++)
                  numberPrefix[j]=Character.getNumericValue(tmp.charAt(j));
               break;
            case 'n':
               numbersToGenerate=Integer.parseInt(g.getOptarg());
               break;
            case 'R':
               printFile(README_FILE);
               System.exit(0);
               break;
            case 'V':
               printFile(VERSION_FILE);
               System.exit(0);
               break;
            case 'A':
               printAlgorithms();
               System.exit(0);
               break;
            case 'L':
               listStates=true;
               break;
            default:
               System.exit(0);
         }
      }                           

      if(args.length-g.getOptind()!=1) {
         printFile(USAGE_FILE);
         System.exit(0);
      }
     
      if(minPartitionLen<=0 || maxWeightsLen<=0 || serialNumberLen<0 ||
         failureTolerance<0 || maxPrintedStates<0 || numbersToGenerate<=0) 
         abort("Option out of range");
    
      if(!generationMode && listStates) 
         abort("-L only works in generation mode.");

      if(generationMode && saveFilename!=null) 
         abort("-s doesn't work in generation mode.");

      if(!generationMode) {
         Laboratory l = new Laboratory(args[g.getOptind()]);
         l.work();
      } else if(listStates) {
         printStates(args[g.getOptind()]);
      } else {
         Factory f = new Factory(args[g.getOptind()]);
         f.work();
      }
         
      System.exit(0);   
   }
}


